@@ -30,6 +30,7 @@
 
 #include <vte/vte.h>
 #include "vtepty-private.h"
+#include "vtetypes.hh"
 
 #include <sys/types.h>
 #include <sys/ioctl.h>
@@ -44,6 +45,7 @@
 #include <sys/syslimits.h>
 #endif
 #include <signal.h>
+#include <pthread.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -54,7 +56,10 @@
 #ifdef HAVE_UTIL_H
 #include <util.h>
 #endif
-#ifdef HAVE_STROPTS_H
+#ifdef HAVE_PTY_H
+#include <pty.h>
+#endif
+#if defined(__sun) && defined(HAVE_STROPTS_H)
 #include <stropts.h>
 #endif
 #include <glib.h>
@@ -63,10 +68,9 @@
 
 #include <glib/gi18n-lib.h>
 
-#if defined(HAVE_PTSNAME_R) || defined(HAVE_PTSNAME) || defined(TIOCGPTN)
-#define HAVE_UNIX98_PTY
-#else
-#undef HAVE_UNIX98_PTY
+/* NSIG isn't in POSIX, so if it doesn't exist use this here. See bug #759196 */
+#ifndef NSIG
+#define NSIG (8 * sizeof(sigset_t))
 #endif
 
 #define VTE_VERSION_NUMERIC ((VTE_MAJOR_VERSION) * 10000 + (VTE_MINOR_VERSION) * 100 + (VTE_MICRO_VERSION))
@@ -77,121 +81,9 @@
 
 #define I_(string) (g_intern_static_string(string))
 
-/* Reset the handlers for all known signals to their defaults.  The parent
- * (or one of the libraries it links to) may have changed one to be ignored. */
-static void
-_vte_pty_reset_signal_handlers(void)
-{
-#ifdef SIGHUP
-	signal(SIGHUP,  SIG_DFL);
-#endif
-	signal(SIGINT,  SIG_DFL);
-	signal(SIGQUIT, SIG_DFL);
-	signal(SIGILL,  SIG_DFL);
-	signal(SIGABRT, SIG_DFL);
-	signal(SIGFPE,  SIG_DFL);
-#ifdef SIGKILL
-	signal(SIGKILL, SIG_DFL);
-#endif
-	signal(SIGSEGV, SIG_DFL);
-#ifdef SIGPIPE
-	signal(SIGPIPE, SIG_DFL);
-#endif
-#ifdef SIGALRM
-	signal(SIGALRM, SIG_DFL);
-#endif
-	signal(SIGTERM, SIG_DFL);
-#ifdef SIGCHLD
-	signal(SIGCHLD, SIG_DFL);
-#endif
-#ifdef SIGCONT
-	signal(SIGCONT, SIG_DFL);
-#endif
-#ifdef SIGSTOP
-	signal(SIGSTOP, SIG_DFL);
-#endif
-#ifdef SIGTSTP
-	signal(SIGTSTP, SIG_DFL);
-#endif
-#ifdef SIGTTIN
-	signal(SIGTTIN, SIG_DFL);
-#endif
-#ifdef SIGTTOU
-	signal(SIGTTOU, SIG_DFL);
-#endif
-#ifdef SIGBUS
-	signal(SIGBUS,  SIG_DFL);
-#endif
-#ifdef SIGPOLL
-	signal(SIGPOLL, SIG_DFL);
-#endif
-#ifdef SIGPROF
-	signal(SIGPROF, SIG_DFL);
-#endif
-#ifdef SIGSYS
-	signal(SIGSYS,  SIG_DFL);
-#endif
-#ifdef SIGTRAP
-	signal(SIGTRAP, SIG_DFL);
-#endif
-#ifdef SIGURG
-	signal(SIGURG,  SIG_DFL);
-#endif
-#ifdef SIGVTALARM
-	signal(SIGVTALARM, SIG_DFL);
-#endif
-#ifdef SIGXCPU
-	signal(SIGXCPU, SIG_DFL);
-#endif
-#ifdef SIGXFSZ
-	signal(SIGXFSZ, SIG_DFL);
-#endif
-#ifdef SIGIOT
-	signal(SIGIOT,  SIG_DFL);
-#endif
-#ifdef SIGEMT
-	signal(SIGEMT,  SIG_DFL);
-#endif
-#ifdef SIGSTKFLT
-	signal(SIGSTKFLT, SIG_DFL);
-#endif
-#ifdef SIGIO
-	signal(SIGIO,   SIG_DFL);
-#endif
-#ifdef SIGCLD
-	signal(SIGCLD,  SIG_DFL);
-#endif
-#ifdef SIGPWR
-	signal(SIGPWR,  SIG_DFL);
-#endif
-#ifdef SIGINFO
-	signal(SIGINFO, SIG_DFL);
-#endif
-#ifdef SIGLOST
-	signal(SIGLOST, SIG_DFL);
-#endif
-#ifdef SIGWINCH
-	signal(SIGWINCH, SIG_DFL);
-#endif
-#ifdef SIGUNUSED
-	signal(SIGUNUSED, SIG_DFL);
-#endif
-}
-
 typedef struct _VtePtyPrivate VtePtyPrivate;
 
-enum TtyOpenMode {
-        TTY_OPEN_BY_NAME,
-        TTY_OPEN_BY_FD
-};
-
 typedef struct {
-        enum TtyOpenMode mode;
-	union {
-		const char *name;
-		int fd;
-	} tty;
-
 	GSpawnChildSetupFunc extra_child_setup;
 	gpointer extra_child_setup_data;
 } VtePtyChildSetupData;
@@ -209,6 +101,7 @@ struct _VtePty {
 struct _VtePtyPrivate {
         VtePtyFlags flags;
         int pty_fd;
+        char *pty_name;
 
         VtePtyChildSetupData child_setup_data;
 
@@ -231,46 +124,53 @@ vte_pty_child_setup (VtePty *pty)
 {
         VtePtyPrivate *priv = pty->priv;
 	VtePtyChildSetupData *data = &priv->child_setup_data;
-	int fd = -1;
-	const char *tty = NULL;
-        char version[7];
 
-        if (priv->foreign) {
-                fd = priv->pty_fd;
-        } else {
-                /* Save the name of the pty -- we'll need it later to acquire
-                * it as our controlling terminal.
-                */
-                switch (data->mode) {
-                        case TTY_OPEN_BY_NAME:
-                                tty = data->tty.name;
-                                break;
-                        case TTY_OPEN_BY_FD:
-                                fd = data->tty.fd;
-                                tty = ttyname(fd);
-                                break;
-                }
-
-                _vte_debug_print (VTE_DEBUG_PTY,
-                                "Setting up child pty: name = %s, fd = %d\n",
-                                        tty ? tty : "(none)", fd);
-
-
-                /* Try to reopen the pty to acquire it as our controlling terminal. */
-                /* FIXMEchpe: why not just use the passed fd in TTY_OPEN_BY_FD mode? */
-                if (tty != NULL) {
-                        int i = open(tty, O_RDWR);
-                        if (i != -1) {
-                                if (fd != -1){
-                                        close(fd);
-                                }
-                                fd = i;
-                        }
-                }
+        /* Unblock all signals */
+        sigset_t set;
+        sigemptyset(&set);
+        if (pthread_sigmask(SIG_SETMASK, &set, nullptr) == -1) {
+                _vte_debug_print(VTE_DEBUG_PTY, "Failed to unblock signals: %m");
+                _exit(127);
         }
 
-	if (fd == -1)
-		_exit (127);
+        /* Reset the handlers for all signals to their defaults.  The parent
+         * (or one of the libraries it links to) may have changed one to be ignored. */
+        for (int n = 1; n < NSIG; n++) {
+                if (n == SIGSTOP || n == SIGKILL)
+                        continue;
+
+                signal(n, SIG_DFL);
+        }
+
+        auto masterfd = priv->pty_fd;
+        if (masterfd == -1)
+                _exit(127);
+
+        if (grantpt(masterfd) != 0) {
+                _vte_debug_print(VTE_DEBUG_PTY, "%s failed: %m", "grantpt");
+                _exit(127);
+        }
+
+	if (unlockpt(masterfd) != 0) {
+                _vte_debug_print(VTE_DEBUG_PTY, "%s failed: %m", "unlockpt");
+                _exit(127);
+        }
+
+	char *name = ptsname(masterfd);
+        if (name == nullptr) {
+		_vte_debug_print(VTE_DEBUG_PTY, "%s failed: %m\n", "ptsname");
+		_exit(127);
+	}
+
+        _vte_debug_print (VTE_DEBUG_PTY,
+                          "Setting up child pty: master FD = %d name = %s\n",
+                          masterfd, name);
+
+        int fd = open(name, O_RDWR);
+        if (fd == -1) {
+                _vte_debug_print (VTE_DEBUG_PTY, "Failed to open PTY: %m\n");
+                _exit(127);
+        }
 
 	/* Start a new session and become process-group leader. */
 #if defined(HAVE_SETSID) && defined(HAVE_SETPGID)
@@ -284,7 +184,7 @@ vte_pty_child_setup (VtePty *pty)
 	ioctl(fd, TIOCSCTTY, fd);
 #endif
 
-#ifdef HAVE_STROPTS_H
+#if defined(__sun) && defined(HAVE_STROPTS_H)
 	if (isastream (fd) == 1) {
 		if ((ioctl(fd, I_FIND, "ptem") == 0) &&
 				(ioctl(fd, I_PUSH, "ptem") == -1)) {
@@ -319,23 +219,19 @@ vte_pty_child_setup (VtePty *pty)
 		}
 	}
 
-	/* Close the original slave descriptor, unless it's one of the stdio
-	 * descriptors. */
+	/* Close the original FD, unless it's one of the stdio descriptors */
 	if (fd != STDIN_FILENO &&
 			fd != STDOUT_FILENO &&
 			fd != STDERR_FILENO) {
 		close(fd);
 	}
 
-	/* Reset our signals -- our parent may have done any number of
-	 * weird things to them. */
-	_vte_pty_reset_signal_handlers();
-
         /* Now set the TERM environment variable */
         /* FIXME: Setting environment here seems to have no effect, the merged envp2 will override on exec.
          * By the way, we'd need to set the one from there, if any. */
         g_setenv("TERM", VTE_DEFAULT_TERM, TRUE);
 
+        char version[7];
         g_snprintf (version, sizeof (version), "%u", VTE_VERSION_NUMERIC);
         g_setenv ("VTE_VERSION", version, TRUE);
 
@@ -398,10 +294,13 @@ __vte_pty_merge_environ (char **envp,
 
         g_hash_table_replace (table, g_strdup ("VTE_VERSION"), g_strdup_printf ("%u", VTE_VERSION_NUMERIC));
 
+	/* Always set this ourself, not allowing replacing from envp */
+	g_hash_table_replace(table, g_strdup("COLORTERM"), g_strdup("truecolor"));
+
 	array = g_ptr_array_sized_new (g_hash_table_size (table) + 1);
         g_hash_table_iter_init(&iter, table);
         while (g_hash_table_iter_next(&iter, (void**) &name, (void**) &value)) {
-                g_ptr_array_add (array, g_strconcat (name, "=", value, NULL));
+                g_ptr_array_add (array, g_strconcat (name, "=", value, nullptr));
         }
         g_assert(g_hash_table_size(table) == array->len);
 	g_hash_table_destroy (table);
@@ -558,7 +457,7 @@ vte_pty_set_size(VtePty *pty,
 			master, columns, rows);
 	ret = ioctl(master, TIOCSWINSZ, &size);
 	if (ret != 0) {
-                int errsv = errno;
+                vte::util::restore_errno errsv;
 
                 g_set_error(error, G_IO_ERROR,
                             g_io_error_from_errno(errsv),
@@ -568,9 +467,6 @@ vte_pty_set_size(VtePty *pty,
 		_vte_debug_print(VTE_DEBUG_PTY,
 				"Failed to set size on %d: %s.\n",
 				master, g_strerror(errsv));
-
-                errno = errsv;
-
                 return FALSE;
 	}
 
@@ -618,7 +514,7 @@ vte_pty_get_size(VtePty *pty,
 				master, size.ws_col, size.ws_row);
                 return TRUE;
 	} else {
-                int errsv = errno;
+                vte::util::restore_errno errsv;
 
                 g_set_error(error, G_IO_ERROR,
                             g_io_error_from_errno(errsv),
@@ -628,270 +524,147 @@ vte_pty_get_size(VtePty *pty,
 		_vte_debug_print(VTE_DEBUG_PTY,
 				"Failed to read size from fd %d: %s\n",
 				master, g_strerror(errsv));
-
-                errno = errsv;
-
                 return FALSE;
 	}
 }
 
-#if defined(HAVE_UNIX98_PTY)
-
-/*
- * _vte_pty_ptsname:
- * @master: file descriptor to the PTY master
- * @error: a location to store a #GError, or %NULL
- *
- * Returns: a newly allocated string containing the file name of the
- *   PTY slave device, or %NULL on failure with @error filled in
- */
-static char *
-_vte_pty_ptsname(int master,
-                 GError **error)
-{
-#if defined(HAVE_PTSNAME_R)
-	gsize len = 1024;
-	char *buf = NULL;
-	int i, errsv;
-	do {
-		buf = (char *) g_malloc0(len);
-		i = ptsname_r(master, buf, len - 1);
-		switch (i) {
-		case 0:
-			/* Return the allocated buffer with the name in it. */
-			_vte_debug_print(VTE_DEBUG_PTY,
-					"PTY slave is `%s'.\n", buf);
-			return buf;
-			break;
-		default:
-                        errsv = errno;
-			g_free(buf);
-                        errno = errsv;
-			buf = NULL;
-			break;
-		}
-		len *= 2;
-	} while ((i != 0) && (errno == ERANGE));
-
-        g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-                    "%s failed: %s", "ptsname_r", g_strerror(errno));
-        return NULL;
-#elif defined(HAVE_PTSNAME)
-	char *p;
-	if ((p = ptsname(master)) != NULL) {
-		_vte_debug_print(VTE_DEBUG_PTY, "PTY slave is `%s'.\n", p);
-		return g_strdup(p);
-	}
-
-        g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-                    "%s failed: %s", "ptsname", g_strerror(errno));
-        return NULL;
-#elif defined(TIOCGPTN)
-	int pty = 0;
-	if (ioctl(master, TIOCGPTN, &pty) == 0) {
-		_vte_debug_print(VTE_DEBUG_PTY,
-				"PTY slave is `/dev/pts/%d'.\n", pty);
-		return g_strdup_printf("/dev/pts/%d", pty);
-	}
-
-        g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-                    "%s failed: %s", "ioctl(TIOCGPTN)", g_strerror(errno));
-        return NULL;
-#else
-#error no ptsname implementation for this platform
-#endif
-}
-
-/*
- * _vte_pty_getpt:
- * @error: a location to store a #GError, or %NULL
- *
- * Opens a file descriptor for the next available PTY master.
- * Sets the descriptor to blocking mode!
- *
- * Returns: a new file descriptor, or %-1 on failure
- */
 static int
-_vte_pty_getpt(GError **error)
+fd_set_cloexec(int fd)
 {
-	int fd, flags, rv;
-#if defined(HAVE_POSIX_OPENPT)
-	fd = posix_openpt(O_RDWR | O_NOCTTY);
-#elif defined(HAVE_GETPT)
-	/* Call the system's function for allocating a pty. */
-	fd = getpt();
-#else
-	/* Try to allocate a pty by accessing the pty master multiplex. */
-	fd = open("/dev/ptmx", O_RDWR | O_NOCTTY);
-	if ((fd == -1) && (errno == ENOENT)) {
-		fd = open("/dev/ptc", O_RDWR | O_NOCTTY); /* AIX */
-	}
-#endif
-        if (fd == -1) {
-                g_set_error (error, VTE_PTY_ERROR,
-                             VTE_PTY_ERROR_PTY98_FAILED,
-                             "%s failed: %s", "getpt", g_strerror(errno));
-                return -1;
-        }
+        int flags = fcntl(fd, F_GETFD, 0);
+        if (flags < 0)
+                return flags;
 
-        rv = fcntl(fd, F_GETFL, 0);
-        if (rv < 0) {
-                int errsv = errno;
-                g_set_error(error, VTE_PTY_ERROR,
-                            VTE_PTY_ERROR_PTY98_FAILED,
-                            "%s failed: %s", "fcntl(F_GETFL)", g_strerror(errno));
-                close(fd);
-                errno = errsv;
-                return -1;
-        }
-
-	/* Set it to blocking. */
-        /* FIXMEchpe: why?? vte_terminal_set_pty does the inverse... */
-        flags = rv & ~(O_NONBLOCK);
-        rv = fcntl(fd, F_SETFL, flags);
-        if (rv < 0) {
-                int errsv = errno;
-                g_set_error(error, VTE_PTY_ERROR,
-                            VTE_PTY_ERROR_PTY98_FAILED,
-                            "%s failed: %s", "fcntl(F_SETFL)", g_strerror(errno));
-                close(fd);
-                errno = errsv;
-                return -1;
-        }
-
-	return fd;
+        return fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
 }
 
-static gboolean
-_vte_pty_grantpt(int master,
-                 GError **error)
+static int
+fd_set_nonblocking(int fd)
 {
-#ifdef HAVE_GRANTPT
-        int rv;
-
-        rv = grantpt(master);
-        if (rv != 0) {
-                int errsv = errno;
-                g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-                            "%s failed: %s", "grantpt", g_strerror(errsv));
-                errno = errsv;
-                return FALSE;
-        }
-#endif
-        return TRUE;
+        int flags = fcntl(fd, F_GETFL, 0);
+        if (flags < 0)
+                return -1;
+        if ((flags & O_NONBLOCK) != 0)
+                return 0;
+        return fcntl(fd, F_SETFL, flags | O_NONBLOCK);
 }
 
-static gboolean
-_vte_pty_unlockpt(int fd,
-                  GError **error)
+static int
+fd_set_cpkt(int fd)
 {
-        int rv;
-#ifdef HAVE_UNLOCKPT
-	rv = unlockpt(fd);
-        if (rv != 0) {
-                int errsv = errno;
-                g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-                            "%s failed: %s", "unlockpt", g_strerror(errsv));
-                errno = errsv;
-                return FALSE;
+        /* tty_ioctl(4) -> every read() gives an extra byte at the beginning
+         * notifying us of stop/start (^S/^Q) events. */
+        int one = 1;
+        return ioctl(fd, TIOCPKT, &one);
+}
+
+static int
+fd_setup(int fd)
+{
+        if (fd_set_cloexec(fd) < 0) {
+                vte::util::restore_errno errsv;
+                _vte_debug_print(VTE_DEBUG_PTY,
+                                 "%s failed: %s", "Setting CLOEXEC flag", g_strerror(errsv));
+                return -1;
         }
-        return TRUE;
-#elif defined(TIOCSPTLCK)
-	int zero = 0;
-	rv = ioctl(fd, TIOCSPTLCK, &zero);
-        if (rv != 0) {
-                int errsv = errno;
-                g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-                            "%s failed: %s", "ioctl(TIOCSPTLCK)", g_strerror(errsv));
-                errno = errsv;
-                return FALSE;
+
+        if (fd_set_nonblocking(fd) < 0) {
+                vte::util::restore_errno errsv;
+                _vte_debug_print(VTE_DEBUG_PTY,
+                                 "%s failed: %s", "Setting O_NONBLOCK flag", g_strerror(errsv));
+                return -1;
         }
-        return TRUE;
-#else
-#error no unlockpt implementation for this platform
-#endif
+
+        if (fd_set_cpkt(fd) < 0) {
+                vte::util::restore_errno errsv;
+                _vte_debug_print(VTE_DEBUG_PTY,
+                                 "%s failed: %s", "ioctl(TIOCPKT)", g_strerror(errsv));
+                return -1;
+        }
+
+        return 0;
 }
 
 /*
- * _vte_pty_open_unix98:
+ * _vte_pty_open_posix:
  * @pty: a #VtePty
  * @error: a location to store a #GError, or %NULL
  *
  * Opens a new file descriptor to a new PTY master.
  *
- * Returns: %TRUE on success, %FALSE on failure with @error filled in
+ * Returns: the new PTY's master FD, or -1
  */
-static gboolean
-_vte_pty_open_unix98(VtePty *pty,
-                     GError **error)
+static int
+_vte_pty_open_posix(void)
 {
-        VtePtyPrivate *priv = pty->priv;
-	int fd;
-	char *buf;
-
 	/* Attempt to open the master. */
-	fd = _vte_pty_getpt(error);
-	if (fd == -1)
-                return FALSE;
+        vte::util::smart_fd fd;
+        fd = posix_openpt(O_RDWR | O_NOCTTY | O_NONBLOCK | O_CLOEXEC);
+#ifndef __linux__
+        /* Other kernels may not support CLOEXEC or NONBLOCK above, so try to fall back */
+        bool need_cloexec = false, need_nonblocking = false;
+        if (fd == -1 && errno == EINVAL) {
+                /* Try without NONBLOCK and apply the flag afterward */
+                need_nonblocking = true;
+                fd = posix_openpt(O_RDWR | O_NOCTTY | O_CLOEXEC);
+                if (fd == -1 && errno == EINVAL) {
+                        /* Try without CLOEXEC and apply the flag afterwards */
+                        need_cloexec = true;
+                        fd = posix_openpt(O_RDWR | O_NOCTTY);
+                }
+        }
+#endif /* !linux */
 
-	_vte_debug_print(VTE_DEBUG_PTY, "Allocated pty on fd %d.\n", fd);
-
-        /* Read the slave number and unlock it. */
-        if ((buf = _vte_pty_ptsname(fd, error)) == NULL ||
-            !_vte_pty_grantpt(fd, error) ||
-            !_vte_pty_unlockpt(fd, error)) {
-                int errsv = errno;
+        if (fd == -1) {
+                vte::util::restore_errno errsv;
                 _vte_debug_print(VTE_DEBUG_PTY,
-                                "PTY setup failed, bailing.\n");
-                close(fd);
-                errno = errsv;
-                return FALSE;
+                                 "%s failed: %s", "posix_openpt", g_strerror(errsv));
+                return -1;
         }
 
-        priv->pty_fd = fd;
-        priv->child_setup_data.mode = TTY_OPEN_BY_NAME;
-        priv->child_setup_data.tty.name = buf;
+#ifndef __linux__
+        if (need_cloexec && fd_set_cloexec(fd) < 0) {
+                vte::util::restore_errno errsv;
+                _vte_debug_print(VTE_DEBUG_PTY,
+                                 "%s failed: %s", "Setting CLOEXEC flag", g_strerror(errsv));
+                return -1;
+        }
 
-        return TRUE;
+        if (need_nonblocking && fd_set_nonblocking(fd) < 0) {
+                vte::util::restore_errno errsv;
+                _vte_debug_print(VTE_DEBUG_PTY,
+                                 "%s failed: %s", "Setting NONBLOCK flag", g_strerror(errsv));
+                return -1;
+        }
+#endif /* !linux */
+
+        if (fd_set_cpkt(fd) < 0) {
+                vte::util::restore_errno errsv;
+                _vte_debug_print(VTE_DEBUG_PTY,
+                                 "%s failed: %s", "ioctl(TIOCPKT)", g_strerror(errsv));
+                return -1;
+        }
+
+	_vte_debug_print(VTE_DEBUG_PTY, "Allocated pty on fd %d.\n", (int)fd);
+
+        return fd.steal();
 }
 
-#elif defined(HAVE_OPENPTY)
-
-/*
- * _vte_pty_open_bsd:
- * @pty: a #VtePty
- * @error: a location to store a #GError, or %NULL
- *
- * Opens new file descriptors to a new PTY master and slave.
- *
- * Returns: %TRUE on success, %FALSE on failure with @error filled in
- */
-static gboolean
-_vte_pty_open_bsd(VtePty *pty,
-                  GError **error)
+static int
+_vte_pty_open_foreign(int masterfd /* consumed */)
 {
-	VtePtyPrivate *priv = pty->priv;
-	int parentfd, childfd;
+        vte::util::smart_fd fd(masterfd);
+        if (fd == -1) {
+                errno = EBADF;
+                return -1;
+        }
 
-	if (openpty(&parentfd, &childfd, NULL, NULL, NULL) != 0) {
-		int errsv = errno;
-		g_set_error(error, VTE_PTY_ERROR, VTE_PTY_ERROR_PTY98_FAILED,
-			    "%s failed: %s", "openpty", g_strerror(errsv));
-		errno = errsv;
-		return FALSE;
-	}
+        if (fd_setup(fd) < 0)
+                return -1;
 
-	priv->pty_fd = parentfd;
-	priv->child_setup_data.mode = TTY_OPEN_BY_FD;
-	priv->child_setup_data.tty.fd = childfd;
-
-	return TRUE;
+        return fd.steal();
 }
 
-#else
-#error Have neither UNIX98 PTY nor BSD openpty!
-#endif /* HAVE_UNIX98_PTY */
-
 /**
  * vte_pty_set_utf8:
  * @pty: a #VtePty
@@ -920,10 +693,9 @@ vte_pty_set_utf8(VtePty *pty,
         g_return_val_if_fail (priv->pty_fd != -1, FALSE);
 
         if (tcgetattr(priv->pty_fd, &tio) == -1) {
-                int errsv = errno;
+                vte::util::restore_errno errsv;
                 g_set_error(error, G_IO_ERROR, g_io_error_from_errno(errsv),
                             "%s failed: %s", "tcgetattr", g_strerror(errsv));
-                errno = errsv;
                 return FALSE;
         }
 
@@ -937,10 +709,9 @@ vte_pty_set_utf8(VtePty *pty,
         /* Only set the flag if it changes */
         if (saved_cflag != tio.c_iflag &&
             tcsetattr(priv->pty_fd, TCSANOW, &tio) == -1) {
-                int errsv = errno;
+                vte::util::restore_errno errsv;
                 g_set_error(error, G_IO_ERROR, g_io_error_from_errno(errsv),
                             "%s failed: %s", "tcgetattr", g_strerror(errsv));
-                errno = errsv;
                 return FALSE;
 	}
 #endif
@@ -978,7 +749,6 @@ vte_pty_initable_init (GInitable *initable,
 {
         VtePty *pty = VTE_PTY (initable);
         VtePtyPrivate *priv = pty->priv;
-        gboolean ret = FALSE;
 
         if (cancellable != NULL) {
                 g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
@@ -986,25 +756,20 @@ vte_pty_initable_init (GInitable *initable,
                 return FALSE;
         }
 
-        /* If we already have a (foreign) FD, we're done. */
         if (priv->foreign) {
-                g_assert(priv->pty_fd != -1);
-                return TRUE;
+                priv->pty_fd = _vte_pty_open_foreign(priv->pty_fd);
+        } else {
+                priv->pty_fd = _vte_pty_open_posix();
         }
 
-#if defined(HAVE_UNIX98_PTY)
-        ret = _vte_pty_open_unix98(pty, error);
-#elif defined(HAVE_OPENPTY)
-        ret = _vte_pty_open_bsd(pty, error);
-#else
-#error Have neither UNIX98 PTY nor BSD openpty!
-#endif
+        if (priv->pty_fd == -1) {
+                vte::util::restore_errno errsv;
+                g_set_error(error, G_IO_ERROR, g_io_error_from_errno(errsv),
+                            "Failed to open PTY: %s", g_strerror(errsv));
+                return FALSE;
+        }
 
-	_vte_debug_print(VTE_DEBUG_PTY,
-			"vte_pty_initable_init returning %s with ptyfd = %d\n",
-			ret ? "TRUE" : "FALSE", priv->pty_fd);
-
-	return ret;
+        return TRUE;
 }
 
 static void
@@ -1016,6 +781,7 @@ vte_pty_initable_iface_init (GInitableIface  *iface)
 /* GObjectClass impl */
 
 G_DEFINE_TYPE_WITH_CODE (VtePty, vte_pty, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (VtePty)
                          G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, vte_pty_initable_iface_init))
 
 static void
@@ -1023,7 +789,7 @@ vte_pty_init (VtePty *pty)
 {
         VtePtyPrivate *priv;
 
-        priv = pty->priv = G_TYPE_INSTANCE_GET_PRIVATE (pty, VTE_TYPE_PTY, VtePtyPrivate);
+        priv = pty->priv = (VtePtyPrivate *)vte_pty_get_instance_private (pty);
 
         priv->flags = VTE_PTY_DEFAULT;
         priv->pty_fd = -1;
@@ -1036,12 +802,6 @@ vte_pty_finalize (GObject *object)
         VtePty *pty = VTE_PTY (object);
         VtePtyPrivate *priv = pty->priv;
 
-        if (priv->child_setup_data.mode == TTY_OPEN_BY_FD &&
-            priv->child_setup_data.tty.fd != -1) {
-                /* Close the child FD */
-                close(priv->child_setup_data.tty.fd);
-        }
-
         /* Close the master FD */
         if (priv->pty_fd != -1) {
                 close(priv->pty_fd);
@@ -1102,8 +862,6 @@ vte_pty_class_init (VtePtyClass *klass)
 {
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-        g_type_class_add_private(object_class, sizeof(VtePtyPrivate));
-
         object_class->set_property = vte_pty_set_property;
         object_class->get_property = vte_pty_get_property;
         object_class->finalize     = vte_pty_finalize;
